from __future__ import absolute_import
from __future__ import print_function
import collections
import sys
from xml.parsers import expat
from . import Extractor
from . import Message


def _open(filename):
    """Injection point for tests."""
    return open(filename, "rb")


class ZCMLExtractor(Extractor):
    """Zope Configuration Markup Language (ZCML)"""

    extensions = [".zcml"]
    ATTRIBUTES = set(["title", "description"])

    def __call__(self, filename, options, fileobj=None, lineno=0):
        self.filename = filename
        self.target_domain = options.domain
        self.messages = []
        self.parser = expat.ParserCreate()
        self.parser.StartElementHandler = self.StartElementHandler
        self.parser.EndElementHandler = self.EndElementHandler
        self.domainstack = collections.deque()
        if fileobj is None:
            fileobj = _open(filename)
        try:
            self.parser.ParseFile(fileobj)
        except expat.ExpatError as e:
            print(
                "Aborting due to parse error in %s: %s" % (filename, e), file=sys.stderr
            )
            sys.exit(1)
        return self.messages

    def add_message(self, msgid):
        self.messages.append(
            Message(
                None,
                msgid,
                None,
                [],
                "",
                "",
                (self.filename, (self.parser.CurrentLineNumber)),
            )
        )

    def StartElementHandler(self, name, attributes):
        if "i18n_domain" in attributes:
            self.domainstack.append(attributes["i18n_domain"])
        elif self.domainstack:
            self.domainstack.append(self.domainstack[-1])

        if not self.domainstack:
            return

        if self.target_domain in [None, self.domainstack[-1]]:
            for (key, value) in attributes.items():
                if key in self.ATTRIBUTES:
                    self.add_message(value)

    def EndElementHandler(self, name):
        if self.domainstack:
            self.domainstack.pop()



from .compat import EntryPointLoadError
from .compat import iter_entry_points
from .compat import load_entry_point
from .python import KEYWORDS
from .python import parse_keyword
from . import EXTRACTORS
from . import Message
from . import check_c_format
from . import check_python_format
from . import Extractor
from . import update_keywords


class BabelExtractor(Extractor):
    extensions = []
    extractor = None
    default_config = {
        "comment-tags": "",
    }

    def __call__(self, filename, options, fileobj=None, firstline=0):
        self.keywords = KEYWORDS.copy()
        update_keywords(self.keywords, options.keywords)
        if fileobj is None:
            fileobj = open(filename, "rb")
        comment_tags = self.config["comment-tags"].split()
        messages = self.extractor(
            fileobj, list(self.keywords.keys()), comment_tags, self.config
        )
        for (lineno, function, args, comment) in messages:
            if not isinstance(args, (list, tuple)):
                args = [args]
            if function in self.keywords:
                args = [(None, a, lineno) for a in args]
                (domain, msgctxt, msgid, msgid_plural, c) = parse_keyword(
                    args, self.keywords[function], filename, lineno
                )
                if c:
                    comment.append(c)
            else:
                msgid = args[0]
                domain = msgctxt = msgid_plural = None

            if domain and self.options.domain and domain != self.options.domain:
                continue
            comment = " ".join(comment)
            flags = []
            check_c_format(msgid, flags)
            check_python_format(msgid, flags)
            yield Message(
                msgctxt,
                msgid,
                msgid_plural,
                flags,
                comment,
                "",
                (filename, firstline + lineno),
            )


def register_babel_plugins():
    for entry_point in iter_entry_points("babel.extractors"):
        name = entry_point.name
        try:
            extractor = load_entry_point(entry_point)
        except EntryPointLoadError:
            # skip this entry point since at least one required dependency can
            # not be found
            extractor = None
        if extractor:
            cls = type(
                "BabelExtractor_%s" % name,
                (BabelExtractor, object),
                {
                    "extractor": staticmethod(extractor),
                    "__doc__": extractor.__doc__.splitlines()[0],
                },
            )
            EXTRACTORS["babel-%s" % name] = cls()



import click
import collections
import textwrap
import polib


def verify_po(path, show_path):
    leader = "[%s] " % path if show_path else ""
    try:
        catalog = polib.pofile(path)
    except UnicodeDecodeError:
        click.echo("Character encoding problems occured while parsing %s" % path)
        click.echo("Perhaps this is not a PO file?")
        return
    msgids = collections.defaultdict(int)
    reverse_map = collections.defaultdict(list)

    for entry in catalog:
        key = (entry.msgctxt, entry.msgid)
        msgids[key] += 1
        if entry.msgstr:
            reverse_map[entry.msgstr].append(key)

    for (key, count) in msgids.items():
        if count == 1:
            continue
        click.echo("%sMessage repeated %d times:" % (leader, count))
        (context, msgid) = key
        if context:
            msgid = "[%s] %s" % (context, msgid)
        click.echo(
            textwrap.fill(msgid, initial_indent=" " * 5, subsequent_indent=" " * 8)
        )
        click.echo()

    for (msgstr, keys) in reverse_map.items():
        if len(keys) == 1:
            continue

        click.echo("%sTranslation:" % leader)
        click.echo(
            textwrap.fill(msgstr, initial_indent=" " * 8, subsequent_indent=" " * 8)
        )
        click.echo("Used for %d canonical texts:" % len(keys))
        for (idx, info) in enumerate(keys):
            (context, msgid) = info
            if context:
                msgid = "[%s] %s" % (context, msgid)
            click.echo(
                textwrap.fill(
                    msgid, initial_indent="%-8d" % (idx + 1), subsequent_indent=8 * " "
                )
            )
        click.echo()


@click.command()
@click.argument("input", nargs=-1, type=click.Path(exists=True), metavar="PO-file")
def main(input):
    "Perform sanity checks on PO files"

    show_path = len(input) > 1
    for path in input:
        verify_po(path, show_path)



import io

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from operator import attrgetter
import os
import re
import sys
import tempfile
import time

try:
    from configparser import SafeConfigParser
except ImportError:
    try:
        from configparser import ConfigParser as SafeConfigParser
    except ImportError:
        from ConfigParser import SafeConfigParser
import click
import polib
from lingua.extractors import get_extractor
from lingua.extractors import register_extractors
from lingua.extractors.babel import register_babel_plugins
from lingua.extractors import EXTRACTORS
from lingua.extractors import EXTENSIONS
from lingua import __version__


def po_timestamp():
    local = time.localtime()
    offset = -(time.altzone if local.tm_isdst else time.timezone)
    return "%s%s%s" % (
        time.strftime("%Y-%m-%d %H:%M", local),
        "-" if offset < 0 else "+",
        time.strftime("%H%M", time.gmtime(abs(offset))),
    )


def _same_text(a, b):
    a = re.sub(r"\s+", " ", a)
    b = re.sub(r"\s+", " ", b)
    return a == b


class POEntry(polib.POEntry):
    def __init__(self, *a, **kw):
        polib.POEntry.__init__(self, *a, **kw)
        self._comments = []
        self._tcomments = []

    @property
    def comment(self):
        return "\n".join(self._comments)

    @comment.setter
    def comment(self, value):
        pass

    @property
    def tcomment(self):
        return "\n".join(self._tcomments)

    @tcomment.setter
    def tcomment(self, value):
        pass

    def __eq__(self, other):
        r = super(POEntry, self).__eq__(other)
        if not r:
            return False
        return _same_text(other.comment, self.comment) and _same_text(
            other.tcomment, self.tcomment
        )

    def update(self, message, add_occurrences=True):
        if add_occurrences:
            self.occurrences.append((message.location[0], str(message.location[1])))
        self.flags.extend(f for f in message.flags if f not in self.flags)
        if message.comment not in self._comments:
            self._comments.append(message.comment)
        if message.tcomment not in self._tcomments:
            self._tcomments.append(message.tcomment)


class POFile(polib.POFile):
    copyright = None
    package_name = None

    def metadata_as_entry(self):
        entry = polib.POFile.metadata_as_entry(self)
        year = time.localtime().tm_year
        header = ["SOME DESCRIPTIVE TITLE"]
        if self.copyright_holder:
            header.append("Copyright (C) %d %s" % (year, self.copyright_holder))
        header.append(
            "This file is distributed under the same license as the %s package."
            % self.package_name
        )
        header.append("FIRST AUTHOR <EMAIL@ADDRESS>, %d." % year)
        entry.tcomment = "\n".join(header)
        return entry


def no_duplicates(iterator):
    seen = set()
    for item in iterator:
        if item in seen:
            continue
        seen.add(item)
        yield item


def list_files(files_from, sources):
    if files_from:
        for filename in files_from:
            if filename.startswith("#") or not filename.strip():
                continue
            yield filename.rstrip()
    for file in sources:
        if os.path.isfile(file):
            yield file
        elif os.path.isdir(file):
            for (dirpath, dirnames, filenames) in os.walk(file):
                for file in filenames:
                    if get_extractor(file) is not None:
                        yield os.path.join(dirpath, file)
        else:
            click.echo("Invalid file type for %s" % file, err=True)
            sys.exit(1)


def find_file(filename, search_path=[]):
    """Return the filename for a given file, checking search paths."""
    paths = [os.path.curdir] + search_path
    for path in paths:
        filename = os.path.join(path, filename)
        if os.path.isfile(filename):
            return filename
    return None


def strip_linenumbers(entry):
    seen = set()
    occurrences = []
    for location, line in entry.occurrences:
        if location in seen:
            continue
        occurrences.append((location, ""))
        seen.add(location)
    entry.occurrences = occurrences


def create_catalog(
    width, copyright_holder, package_name, package_version, msgid_bugs_address
):
    catalog = POFile(wrapwidth=width)
    catalog.copyright_holder = copyright_holder
    catalog.package_name = package_name
    catalog.metadata_is_fuzzy = True
    catalog.metadata = OrderedDict()
    catalog.metadata["Project-Id-Version"] = " ".join(
        filter(None, [package_name, package_version])
    )
    if msgid_bugs_address:
        catalog.metadata["Report-Msgid-Bugs-To"] = msgid_bugs_address
    catalog.metadata["POT-Creation-Date"] = po_timestamp()
    catalog.metadata["PO-Revision-Date"] = "YEAR-MO-DA HO:MI+ZONE"
    catalog.metadata["Last-Translator"] = "FULL NAME <EMAIL@ADDRESS>"
    catalog.metadata["Language-Team"] = "LANGUAGE <LL@li.org>"
    catalog.metadata["Language"] = "LANGUAGE"
    catalog.metadata["MIME-Version"] = "1.0"
    catalog.metadata["Content-Type"] = "text/plain; charset=UTF-8"
    catalog.metadata["Content-Transfer-Encoding"] = "8bit"
    catalog.metadata["Generated-By"] = "Lingua %s" % __version__
    return catalog


def _register_extension(extension, extractor):
    if extractor not in EXTRACTORS:
        click.echo(
            "Unknown extractor %s. Check --list-extractors for available options"
            % extractor,
            err=True,
        )
        sys.exit(1)
    EXTENSIONS[extension] = extractor


def read_config(cfg_file):
    config = SafeConfigParser()
    config.readfp(cfg_file)
    for section in config.sections():
        if section == "extensions":
            for (extension, extractor) in config.items(section):
                _register_extension(extension, extractor)
        elif section.startswith("extractor:"):
            extractor = section[10:]
            if extractor not in EXTRACTORS:
                click.echo(
                    "Unknown extractor %s. "
                    "Check --list-extractors for available options" % extractor,
                    err=True,
                )
                sys.exit(1)
            extractor_config = dict(config.items(section))
            EXTRACTORS[extractor].update_config(**extractor_config)
        elif section.startswith("extension"):
            click.echo(
                "Use of %s section is obsolete. "
                'Please use the "extensions" section.' % section,
                err=True,
            )
            extension = section[10:]
            plugin = config.get(section, "plugin")
            if not plugin:
                click.echo("No plugin defined for extension %s" % extension, err=True)
            _register_extension(extension, plugin)


def _summarise(catalog):
    summary = {}
    for entry in catalog:
        if entry.obsolete:
            continue
        summary[(entry.msgid, entry.msgctxt)] = entry
    return summary


def identical(a, b):
    """Check if two catalogs are identical, ignoring metadata."""
    a = _summarise(a)
    b = _summarise(b)
    return a == b


def save_catalog(catalog, filename):
    if os.path.exists(filename):
        old_catalog = None
        try:
            old_catalog = polib.pofile(filename)
        except (OSError, UnicodeDecodeError):
            pass
        if old_catalog is not None and identical(catalog, old_catalog):
            click.echo("No changes found - not replacing %s" % filename)
            return
    (fd, tmpfile) = tempfile.mkstemp(dir=os.path.dirname(filename), text=True)
    output = io.open(fd, "wt", encoding=catalog.encoding)
    output.write(catalog.__unicode__())
    output.close()
    os.rename(tmpfile, filename)


def _location_sort_key(msg):
    locations = [(fn, int(line)) for (fn, line) in msg.occurrences]
    locations.sort()  # Sort so first occurence is always used.
    return locations


class ExtractorOptions:
    def __init__(self, comment_tag, domain, keywords):
        self.comment_tag = comment_tag
        self.domain = domain
        self.keywords = keywords


@click.command()
@click.option(
    "-c",
    "--config",
    "cfg_file",
    metavar="CONFIG",
    help="Read configuration from CONFIG file",
    type=click.File(),
)
# Input options
@click.option(
    "-f",
    "--files-from",
    metavar="FILE",
    type=click.File(),
    help="Get list of files to process from FILE",
)
@click.option(
    "-D",
    "--directory",
    metavar="DIRECTORY",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    multiple=True,
    help="Add DIRECTORY to list of paths to check for input files",
)
@click.argument("sources", nargs=-1, type=click.Path(exists=True))
@click.option(
    "--list-extractors", is_flag=True, help="List all known extraction plugins"
)
# Output options
@click.option(
    "-o",
    "--output",
    metavar="FILE",
    type=click.Path(exists=False, dir_okay=False, writable=True),
    default="messages.pot",
    help="Filename for generated POT file",
)
@click.option(
    "--add-location/--no-location",
    "location",
    default=True,
    help="Include location information",
)
@click.option(
    "--linenumbers/--no-linenumbers",
    default=True,
    help="Include line numbers in location information",
)
@click.option("-w", "--width", metavar="NUMBER", default=79, help="Output width")
@click.option(
    "-s",
    "--sort-output",
    "sort_order",  # babel compatibility
    flag_value="msgid",
    help="Order messages by their msgid",
)
@click.option(
    "-F",
    "--sort-by-file",
    "sort_order",
    flag_value="location",
    help="Order messages by file location",
)
# Extraction configuration
@click.option("-d", "--domain", help="Domain to extract")
@click.option(
    "-k",
    "--keyword",
    "keywords",
    metavar="WORD",
    multiple=True,
    help="Look for WORD as additional keyword",
)
@click.option(
    "-C",
    "--add-comments",
    "comment_tag",
    metavar="TAG",
    help="Add comments prefixed by TAG to messages, or all if no tag is given",
)
# POT metadata
@click.option(
    "--copyright-holder",
    metavar="STRING",
    help="Specifies the copyright holder for the texts",
)
@click.option(
    "--package-name",
    metavar="NAME",
    default="PACKAGE",
    help="Package name to use in the generated POT file",
)
@click.option(
    "--package-version",
    metavar="Version",
    default="1.0",
    help="Package version to use in the generated POT file",
)
@click.option(
    "--msgid-bugs-address", metavar="EMAIL", help="Email address bugs should be send to"
)
def main(
    cfg_file,
    files_from,
    directory,
    sources,
    list_extractors,
    output,
    location,
    linenumbers,
    width,
    sort_order,
    domain,
    keywords,
    comment_tag,
    copyright_holder,
    package_name,
    package_version,
    msgid_bugs_address,
):
    "Extract translatable strings."
    directory = list(directory)
    register_extractors()
    register_babel_plugins()

    if comment_tag is None:
        comment_tag = True
    if list_extractors:
        for extractor in sorted(EXTRACTORS):
            click.echo("%-17s %s" % (extractor, EXTRACTORS[extractor].__doc__ or ""))
        return

    if cfg_file:
        read_config(cfg_file)
    else:
        user_home = os.path.expanduser("~")
        global_config = os.path.join(user_home, ".config", "lingua")
        if os.path.exists(global_config):
            read_config(open(global_config, "r"))

    catalog = create_catalog(
        width, copyright_holder, package_name, package_version, msgid_bugs_address
    )

    scanned = 0
    for filename in no_duplicates(list_files(files_from, sources)):
        real_filename = find_file(filename, directory)
        if real_filename is None:
            click.echo("Can not find file %s" % filename, err=True)
            sys.exit(1)
        extractor = get_extractor(real_filename)
        if extractor is None:
            click.echo("No extractor available for file %s" % filename, err=True)
            sys.exit(1)

        extractor_options = ExtractorOptions(
            comment_tag=comment_tag,
            domain=domain,
            keywords=keywords,
        )
        for message in extractor(real_filename, extractor_options):
            entry = catalog.find(message.msgid, msgctxt=message.msgctxt)
            if entry is None:
                entry = POEntry(msgctxt=message.msgctxt, msgid=message.msgid)
                if message.msgid_plural:
                    entry.msgid_plural = message.msgid_plural
                    entry.msgstr_plural[0] = ""
                    entry.msgstr_plural[1] = ""
                catalog.append(entry)
            entry.update(message, add_occurrences=location)
        scanned += 1
    if not scanned:
        click.echo("No files scanned, aborting", err=True)
        sys.exit(1)
    if not catalog:
        click.echo("No translatable strings found, aborting", err=True)
        sys.exit(2)

    if sort_order == "msgid":
        catalog.sort(key=attrgetter("msgid"))
    elif sort_order == "location":
        catalog.sort(key=_location_sort_key)

    if not linenumbers:
        for entry in catalog:
            strip_linenumbers(entry)

    save_catalog(catalog, output)


if __name__ == "__main__":
    main()



from __future__ import print_function
import abc
import collections
import os
import re
import sys
from .compat import add_metaclass
from .compat import EntryPointLoadError
from .compat import iter_entry_points
from .compat import load_entry_point


Message = collections.namedtuple(
    "Message", "msgctxt msgid msgid_plural flags comment tcomment location"
)

EXTRACTORS = {}
EXTENSIONS = {}


def get_extractor(filename):
    ext = os.path.splitext(filename)[1]
    try:
        return EXTRACTORS[EXTENSIONS[ext]]
    except KeyError:
        return None


# Based on http://www.cplusplus.com/reference/cstdio/printf/
# Note that we skip the space-flag in this list, since this creates too
# many false positives.
_C_FORMAT = re.compile(
    r"""
        %
        [+#0-]?               # flags
        (\d+|\*)?             # width
        (\.(\d+|\*))?         # precision
        (hh?|ll?|j|z|t|L)?    # length
        [diuoxXfFeEgGaAcspn%] # specifier
        """,
    re.VERBOSE,
)


def check_c_format(buf, flags):
    if "no-c-format" in flags or "c-format" in flags:
        return
    formats = list(re.finditer("%(?!%)", buf))
    if formats and all(_C_FORMAT.match(buf[m.start() :]) is not None for m in formats):
        flags.append("c-format")


# Based on http://docs.python.org/2/library/string.html#format-string-syntax
_PYTHON_FORMAT = re.compile(
    r"""
        \{
            (([_A-Za-z](\w*)(\.[_a-z]\w*|\[\d+\])?)|\w+)?  # fieldname
            (![rs])?  # conversion
            (:\.?[<>=^]?[+ -]?\w*,?(\.\w+)?[bcdeEfFgGnosxX%]?)?  # format_spec
        \}
        """,
    re.VERBOSE,
)


def check_python_format(buf, flags):
    if "no-python-format" in flags or "python-format" in flags:
        return
    if _PYTHON_FORMAT.search(buf) is not None:
        flags.append("python-format")


def check_comment_flags(comment):
    flags = re.match("\\[\\s*(.*?)\\s*\\]\\s*(.*)", comment)
    if flags is not None:
        return (re.split("\\s*,\\s*", flags.group(1)), flags.group(2))
    else:
        return ([], comment)


class Keyword(object):
    msgctxt_param = None
    domain_param = None
    comment = ""
    required_arguments = None

    _comment_arg = re.compile(r'^"(.*)"$')

    def __init__(
        self,
        function,
        msgid_param=1,
        msgid_plural_param=None,
        domain_param=None,
        msgctxt_param=None,
    ):
        self.function = function
        self.msgid_param = msgid_param
        self.msgid_plural_param = msgid_plural_param
        self.domain_param = domain_param
        self.msgctxt_param = msgctxt_param

    @classmethod
    def from_spec(cls, spec):
        if ":" not in spec:
            return cls(spec)
        try:
            (function, args) = spec.split(":", 1)
            kw = cls(function)
            seen_msgid_param = False
            while args:
                if cls._comment_arg.match(args) is not None:
                    kw.comment = args[1:-1]
                    break
                (param, args) = args.split(",", 1) if "," in args else (args, "")
                if param.endswith("c"):
                    kw.msgctxt_param = int(param[:-1])
                elif param.endswith("d"):
                    kw.domain_param = int(param[:-1])
                elif param.endswith("t"):
                    kw.required_arguments = int(param[:-1])
                elif not seen_msgid_param:
                    kw.msgid_param = int(param)
                    seen_msgid_param = True
                else:
                    kw.msgid_plural_param = int(param)
        except SyntaxError:
            raise ValueError("Invalid keyword spec: %s" % spec)
        return kw


def update_keywords(keywords, specs):
    for spec in specs:
        if not spec:
            keywords.clear()
        try:
            kw = Keyword.from_spec(spec)
        except ValueError as e:
            print(e, file=sys.stderr)
            sys.exit(1)
        keywords[kw.function] = kw


@add_metaclass(abc.ABCMeta)
class Extractor(object):
    default_config = {}

    def __init__(self, config=None):
        self.config = self.default_config.copy()
        if config:
            self.config.update(config)

    def update_config(self, **kw):
        self.config.update(kw)

    @abc.abstractproperty
    def extensions(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def __call__(self, filename, options, fileobj=None, lineno=0):
        raise NotImplementedError()


def register_extractors():
    for entry_point in iter_entry_points("lingua.extractors"):
        try:
            extractor = load_entry_point(entry_point)
        except EntryPointLoadError:
            # skip this entry point since at least one required dependency can
            # not be found
            extractor = None
        if extractor:
            if not issubclass(extractor, Extractor):
                raise ValueError("Registered extractor must derive from ``Extractor``")
            EXTRACTORS[entry_point.name] = extractor()
            for extension in extractor.extensions:
                EXTENSIONS[extension] = entry_point.name



from __future__ import print_function
import ast
import io
import sys
import tokenize
import warnings
from . import Extractor
from . import Message
from . import check_comment_flags
from . import check_c_format
from . import check_python_format
from . import Keyword
from . import update_keywords


try:
    basestring
except NameError:
    basestring = str


KEYWORDS = {
    "gettext": Keyword("gettext"),
    "ugettext": Keyword("ugettext"),
    "dgettext": Keyword("dgettext", 2, domain_param=1),
    "ldgettext": Keyword("ldgettext", 2, domain_param=1),
    "ngettext": Keyword("ngettext", 1, 2),
    "lngettext": Keyword("ngettext", 1, 2),
    "ungettext": Keyword("ungettext", 1, 2),
    "dngettext": Keyword("dngettext", 2, 3, domain_param=1),
    "ldngettext": Keyword("dngettext", 2, 3, domain_param=1),
    "pgettext": Keyword("pgettext", 2, msgctxt_param=1),
}


class ParseError(ValueError):
    def __init__(self, msg, lineno):
        self.lineno = lineno
        ValueError.__init__(self, msg)


def parse_keyword(arguments, keyword, filename, firstline):
    # This assumes kw arguments are not used.
    if keyword.required_arguments and len(arguments) != keyword.required_arguments:
        return None

    def get_string(param, error_msg):
        if not param:
            return None
        (arg_name, arg_value, lineno) = arguments[param - 1]
        if not isinstance(arg_value, basestring):
            print(
                "%s[%d]: %s" % (filename, firstline + lineno, error_msg),
                file=sys.stderr,
            )
            raise IndexError()
        else:
            return arg_value

    try:
        domain = get_string(keyword.domain_param, "Domain argument must be a string")
        msgid = get_string(keyword.msgid_param, "Message argument must be a string")
        msgid_plural = get_string(
            keyword.msgid_plural_param, "Plural message argument must be a string"
        )
        msgctxt = get_string(keyword.msgctxt_param, "Context argument must be a string")
        comment = keyword.comment
    except IndexError:
        return None
    return (domain, msgctxt, msgid, msgid_plural, comment)


def parse_translationstring(arguments, filename, firstline):
    if not arguments:
        return None

    msgid = None
    context = None
    default = ""
    args = [a[1] for a in arguments if a[0] is None]
    kwargs = dict((a[0], a[1]) for a in arguments if a[0] is not None)

    if isinstance(args[0], basestring):
        msgid = args[0]
    if len(args) > 2 and isinstance(args[2], basestring):
        default = args[2]
    for (key, value) in kwargs.items():
        if not isinstance(value, basestring):
            continue
        if key == "msgid":
            msgid = value
        elif key == "default":
            default = value
        elif key == "context":
            context = value
    if not msgid:
        return None

    comment = "Default: %s" % default if default else ""
    return (None, context, msgid, None, comment)


def _open(filename):
    """Injection point for tests."""
    return open(filename, "r")


def safe_eval(s):
    if isinstance(s, bytes):
        s = s.decode("utf-8")
    return ast.literal_eval(s)


DYNAMIC = []


class _SafeReadline(object):
    def __init__(self, readline):
        self.readline = readline

    def __call__(self):
        line = self.readline()
        if isinstance(line, bytes):
            warnings.warn(
                "Python extractor called with bytes input. "
                "Please update your plugin to submit unicode instead.",
                UnicodeWarning,
                stacklevel=6,
            )
            line = line.decode("utf-8")
        return line


class TokenStreamer(object):
    def __init__(self, readline):
        self.queue = tokenize.generate_tokens(_SafeReadline(readline))
        self.pushed = []
        self.complete = False

    def __iter__(self):
        return self

    def _transform(self, raw_token):
        (token_type, token, loc_start, loc_end, line) = raw_token
        return (token_type, token, loc_start, self)

    def push(self, token):
        self.pushed.append(token)
        self.complete = False

    def peek(self):
        token = self.next()
        self.push(token)
        return token

    def next(self):
        if self.complete:
            raise StopIteration()
        if self.pushed:
            token = self.pushed.pop()
        else:
            token = self._transform(next(self.queue))
        if token[0] == tokenize.ENDMARKER:
            self.complete = True
        return token

    __next__ = next  # For Python 3 compatibility


class PythonParser(object):
    last_comment = (-2, None)

    def __call__(self, token_stream, options, filename, firstline):
        self.options = options
        if options.comment_tag is True:
            self.include_comments = "all"
        elif options.comment_tag is None:
            self.include_comments = "none"
        else:
            self.include_comments = "tagged"
            self.comment_marker = options.comment_tag
        self.filename = filename
        self.firstline = firstline
        self.messages = []
        self.handler = self.state_skip
        try:
            for (token_type, token, location, _) in token_stream:
                self.process_token(token_type, token, location, token_stream)
        except tokenize.TokenError as e:
            print(
                "Aborting due to parse error in %s[%d]: %s"
                % (filename, firstline + e.args[1][0], e.args[0]),
                file=sys.stderr,
            )
            sys.exit(1)
        except ParseError as e:
            print(
                "Aborting due to parse error in %s[%d]: %s"
                % (filename, firstline + e.lineno, e.args[0]),
                file=sys.stderr,
            )
            sys.exit(1)
        return self.messages

    def process_token(self, token_type, token, location, token_stream):
        if token_type == tokenize.COMMENT:
            self.process_comment(token, location)
        elif token_type in [
            tokenize.INDENT,
            tokenize.DEDENT,
            tokenize.NEWLINE,
            tokenize.NL,
        ]:
            return
        else:
            self.handler(token_type, token, location, token_stream)

    def process_comment(self, token, location):
        if self.include_comments == "none":
            return
        comment = token[1:].strip()  # Remove leading hash
        if self.include_comments == "all" or comment.startswith(self.comment_marker):
            if self.include_comments == "tagged":
                comment = comment[len(self.comment_marker) :].strip()
            (flags, comment) = check_comment_flags(comment)
            if self.messages and self.messages[-1].location[1] == (
                self.firstline + location[0]
            ):
                last_message = self.messages[-1]
                # Comment at the end of the line of a keyword call
                new_comment = []
                if self.messages[-1].comment:
                    new_comment.append(last_message.comment)
                new_comment.append(comment)
                self.messages[-1] = Message(
                    last_message.msgctxt,
                    last_message.msgid,
                    last_message.msgid_plural,
                    last_message.flags,
                    "\n".join(new_comment),
                    last_message.tcomment,
                    last_message.location,
                )
            else:
                if self.last_comment[0] == location[0] - 1:
                    comment = self.last_comment[1] + " " + comment
                    flags = flags + self.last_comment[2]
                self.last_comment = (location[0], comment, flags)

    def state_skip(self, token_type, token, location, token_stream):
        """Ignore all input until we see one of our keywords."""
        if token_type == tokenize.NAME and (token in KEYWORDS or token == "_"):
            self.handler = self.state_in_keyword
            self.keyword = KEYWORDS.get(token, None)
            self.lineno = location[0]
        elif token_type == tokenize.NAME and token == "def":
            self.handler = self.state_skip_function_def

    def state_skip_function_def(self, token_type, token, location, token_stream):
        self.handler = self.state_skip

    def state_in_keyword(self, token_type, token, location, token_stream):
        """Check if the keyword is used in a proper function call."""
        if token_type == tokenize.OP and token == "(":
            self.arguments = []
            self.argument_name = None
            self.in_argument = False
            self.in_string = False
            self.handler = self.state_in_keyword_call
        else:
            self.handler = self.state_skip

    def state_in_keyword_call(self, token_type, token, location, token_stream):
        """Collect all keyword parameters."""
        if token_type != tokenize.STRING:
            self.in_string = False
        if token_type == tokenize.OP:
            if token == ")":
                self.process_keyword()
                self.handler = self.state_skip
            elif token == ",":
                if not self.in_argument:
                    raise ParseError("Unexpected )", location[0])
                self.in_argument = False
            elif token == "(":  # Tuple
                self.in_argument = True
                self.skip_iterable("(", ")", token_stream)
                self.add_argument(DYNAMIC, location[0])
            elif token == "{":  # Dictionary
                self.in_argument = True
                self.skip_iterable("{", "}", token_stream)
                self.add_argument(DYNAMIC, location[0])
            elif token == "[":  # Array
                self.in_argument = True
                self.skip_iterable("[", "]", token_stream)
                self.add_argument(DYNAMIC, location[0])
            elif token == ".":
                pass
            else:
                raise ParseError("Unexpected token: %s" % token, location[0])
        elif token_type == tokenize.STRING:
            if self.in_string:
                token = safe_eval(token)
                self.arguments[-1] = (
                    self.arguments[-1][0],
                    self.arguments[-1][1] + token,
                    self.arguments[-1][2],
                )
            else:
                token = safe_eval(token)
                self.add_argument(token, location[0])
                self.in_argument = True
                self.in_string = True
        elif token_type == tokenize.NUMBER:
            if self.in_argument:
                raise ParseError("Unexpected number: %s" % token, location[0])
            self.add_argument(safe_eval(token), location[0])
        elif token_type == tokenize.NAME:
            self.in_argument = True
            (next_token_type, next_token) = token_stream.peek()[:2]
            if next_token_type == tokenize.OP and next_token in ",)":
                # Variable reference
                self.add_argument(DYNAMIC, location[0])
            elif next_token_type == tokenize.OP and next_token == "(":
                # Function call
                next(token_stream)  # Make sure skip_iterable does not see ( again
                self.skip_iterable("(", ")", token_stream)
                self.add_argument(DYNAMIC, location[0])
            elif next_token_type == tokenize.OP and next_token == "=":
                if self.argument_name is not None:
                    raise ParseError("Unexpected token: %s" % token, location[0])
                next(token_stream)  # Pop the equal token
                self.argument_name = token
        else:
            raise ParseError("Unexpected token: %s" % token, location[0])
        # XXX Does this handle trailing comma: func(foo,x bar,)

    def add_argument(self, value, lineno):
        self.arguments.append((self.argument_name, value, lineno))
        self.argument_name = None

    def skip_iterable(self, start_token, end_token, token_stream):
        depth = 1
        for (token_type, token, loc, _) in token_stream:
            if token_type == tokenize.OP:
                if token == start_token:
                    depth += 1
                elif token == end_token:
                    depth -= 1
                    if depth == 0:
                        return
            elif token_type == tokenize.ENDMARKER:
                raise ParseError("Unexpected end of file", loc[0])
            elif token_type == tokenize.DEDENT:
                raise ParseError("Unexpected dedent", loc[0])

    def process_keyword(self):
        if self.keyword is not None:
            msg = parse_keyword(
                self.arguments, self.keyword, self.filename, self.firstline
            )
        else:
            msg = parse_translationstring(self.arguments, self.filename, self.firstline)
        if msg is None:
            return

        if self.options.domain is not None and msg[0] and msg[0] != self.options.domain:
            return

        comments = []
        flags = []
        if msg[4]:
            comments.append(msg[4])
        if self.last_comment[0] >= (self.lineno - 1):
            comments.append(self.last_comment[1])
            for f in self.last_comment[2]:
                if f not in flags:
                    flags.append(f)
        comment = "\n".join(comments)

        check_c_format(msg[2], flags)
        check_python_format(msg[2], flags)
        self.messages.append(
            Message(
                msg[1],
                msg[2],
                msg[3],
                flags,
                comment,
                "",
                (self.filename, self.firstline + self.lineno),
            )
        )


def _extract_python(filename, source, options, firstline=0):
    if isinstance(source, bytes):
        source = source.decode("utf-8")
    fileobj = io.StringIO(source)
    extractor = PythonExtractor()
    return extractor(filename, options, fileobj, firstline)


class PythonExtractor(Extractor):
    """Python sources"""

    extensions = [".py"]

    def __call__(self, filename, options, fileobj=None, lineno=0):
        update_keywords(KEYWORDS, options.keywords)
        if fileobj is None:
            fileobj = _open(filename)
        token_stream = TokenStreamer(fileobj.readline)
        parser = PythonParser()
        return parser(token_stream, options, filename, lineno)



# Python compatibility support code
# This is taken from six

# Entry points compatibility for importlib.metadata vs pkg_resources
try:
    from importlib.metadata import entry_points

    try:
        # Python 3.10+ returns SelectableGroups with select() method
        entry_points(group="test")

        def iter_entry_points(group):
            return entry_points(group=group)
    except TypeError:
        # Python 3.9 returns a dict-like object
        def iter_entry_points(group):
            eps = entry_points()
            return eps.get(group, [])

    # In importlib.metadata, entry_point.load() doesn't have require parameter.
    # Missing dependencies raise ModuleNotFoundError instead of DistributionNotFound.
    EntryPointLoadError = ModuleNotFoundError

    def load_entry_point(entry_point):
        return entry_point.load()

except ImportError:
    from pkg_resources import DistributionNotFound
    from pkg_resources import working_set

    def iter_entry_points(group):
        return working_set.iter_entry_points(group)

    EntryPointLoadError = DistributionNotFound

    def load_entry_point(entry_point):
        return entry_point.load(require=True)


def add_metaclass(metaclass):
    """Class decorator for creating a class with a metaclass."""

    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        orig_vars.pop("__dict__", None)
        orig_vars.pop("__weakref__", None)
        slots = orig_vars.get("__slots__")
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        return metaclass(cls.__name__, cls.__bases__, orig_vars)

    return wrapper
